
#include "fgdreader.h"
#include <cpl_conv.h>
#include <cpl_string.h>
#include <cpl_port.h>
#include <float.h>


#if defined(__GNUC__)
	#define STRCMPI strcasecmp
#elif defined(WIN32)
	#define STRCMPI stricmp
#else
	#define STRCMPI strcmp
#endif

#ifndef ICONV_CPP_CONST
#define ICONV_CPP_CONST ICONV_CONST
#endif

#ifdef HAVE_EXPAT

static void XMLCALL startElementValidateCbk( void *pUserData, const char *pszName, const char **ppszAttr )
{
	GDALFGDReader* poDS = (GDALFGDReader*) pUserData;
	poDS->startElementValidateCbk( pszName, ppszAttr );
}

static void XMLCALL endElementValidateCbk( void *pUserData, const char *pszName )
{
	GDALFGDReader* poDS = (GDALFGDReader*) pUserData;
	poDS->endElementValidateCbk( pszName );
}

static void XMLCALL dataHandlerValidateCbk( void *pUserData, const char *pszData, int nLen )
{
	GDALFGDReader* poDS = (GDALFGDReader*) pUserData;
	poDS->dataHandlerValidateCbk( pszData, nLen );
}

static void XMLCALL startElement( void *pUserData, const char *pszName, const char **ppszAttr )
{
	GDALFGDReader* poDS = (GDALFGDReader*) pUserData;
	poDS->startElement( pszName, ppszAttr );
}

static void XMLCALL endElement( void *pUserData, const char *pszName )
{
	GDALFGDReader* poDS = (GDALFGDReader*) pUserData;
	poDS->endElement( pszName );
}

static void XMLCALL dataHandler( void *pUserData, const char *pszData, int nLen )
{
	GDALFGDReader* poDS = (GDALFGDReader*) pUserData;
	poDS->dataHandler( pszData, nLen );
}

static int convertSJIStoUTF8( void *data, const char *c )
{
	GDALFGDReader *poDS = (GDALFGDReader *)data;
	return poDS->convertSJIStoUTF8( c );
}

static int XML_ShiftJISEncodingHandler( void *encodingHandlerData, const XML_Char *name, XML_Encoding *info )
{
	int i;

	if ( STRCMPI( name, "shift-jis" ) && STRCMPI( name, "shift_jis" ) )
	{
		return 0;
	}

	// ASCII Map
	info->map['\t'] = '\t';
	info->map['\r'] = '\r';
	info->map['\n'] = '\n';

	for ( i = ' '; i <= '~'; i++ )
	{
		info->map[i] = i;
	}

	for ( i = 0x81; i <= 0x9f; i++ )
	{
		info->map[i] = -2;
	}
	for ( i = 0xa1; i <= 0xdf; i++ )
	{
		info->map[i] = i + 0xfec0;
	}
	for ( i = 0xe0; i <= 0xfc; i++ )
	{
		info->map[i] = -2;
	}

	info->data = encodingHandlerData;
	info->convert = convertSJIStoUTF8;
	info->release = NULL;

	return 1;
}
#endif

GDALFGDReader::GDALFGDReader( const char *pszFilename )
{
	m_bOpen = false;

#if defined(HAVE_EXPAT) && defined(HAVE_ICONV)
	m_nPixelsToPass = 0;
	m_nPixelCount = 0;
	m_nPixelReturn = 0;
	m_validity = FGD_VALIDITY_UNKNOWN;
	m_nStartPix = 0;
	m_bAltiReached = false;

	m_nFlags = 0x0000;

	m_dEast = m_dWest = m_dNorth = m_dSouth = FLT_MAX;
	m_nRow = m_nCol = 0;
	m_nLowX = m_nLowY = m_nHighX = m_nHighY = -1;
	m_nStartX = m_nStartY = -1;

	fp = VSIFOpenL( pszFilename, "r" );
	if ( fp == NULL )
		return;

	m_bOpen = true;

	m_nDataHandlerCounter = 0;

	m_Parser = XML_ParserCreate( NULL );
	XML_SetUserData( m_Parser, this );
	XML_SetElementHandler( m_Parser, ::startElementValidateCbk, ::endElementValidateCbk );
	XML_SetCharacterDataHandler( m_Parser, ::dataHandlerValidateCbk );
	XML_SetUnknownEncodingHandler( m_Parser, XML_ShiftJISEncodingHandler, this );

	m_hIconv = iconv_open( "UTF-16BE", "CP932" );

	scanAll();

#endif
}



GDALFGDReader::~GDALFGDReader()
{
#if defined(HAVE_EXPAT) && defined(HAVE_ICONV)
	XML_ParserFree( m_Parser );

	iconv_close( m_hIconv );
	if ( fp )
		VSIFCloseL( fp );

#endif

	m_bOpen = false;
}

#if defined(HAVE_EXPAT) && defined(HAVE_ICONV)

int GDALFGDReader::convertSJIStoUTF8( const char *s )
{
	char *pszFrom, *pszTo;
	char szFrom[2], szTo[BUFSIZ];
	size_t nInByte, nOutByte;
	int nByte;
	unsigned short nChar = 0;
	ICONV_CPP_CONST char *pszSrcBuf = (ICONV_CONST char *)s;

	if ( !m_hIconv )
		return 0;

	memcpy( szFrom, s, 2 );
	pszFrom = szFrom;
	pszTo = szTo;
	nInByte = 2;
	nOutByte = nInByte * 2;
	iconv( m_hIconv, &pszSrcBuf, &nInByte, &pszTo, &nOutByte );
	*pszTo = '\0';

	nByte = pszTo - szTo;
	if ( nByte > 4 ) return 0;
	pszTo = szTo;
	if ( nByte == 4 )
		pszTo += 2;
	nChar = (*pszTo << 8) | (*(pszTo+1) & 0x00ff);

	return (int)nChar;
}


void GDALFGDReader::startElementValidateCbk(const char *pszName, const char **ppszAttr)
{
	m_szTempChar[0] = '\0';
	m_nTempChar = 0;

	if ( !STRCMPI( pszName, "GI" ) )
	{
		int i;
		m_nFlags |= TAG_GI;
		for ( i = 0; ppszAttr[i] != NULL; i += 2 )
		{
			if ( !strcmp( ppszAttr[i], "version" ) )
			{
				m_pszVersion = CPLStrdup( ppszAttr[i+1] );
				return;
			}
		}
	}
	else if ( !STRCMPI( pszName, "dataset" ) )
	{
		m_nFlags |= TAG_DATASET;
		if ( !( m_nFlags & TAG_GI ) )
		{
			m_validity = FGD_VALIDITY_INVALID;
		}
	}
	else if ( !STRCMPI( pszName, "FGDFeature" ) ||
		 !STRCMPI( pszName, "GCP" ) ||
		 !STRCMPI( pszName, "ElevPt" ) ||
		 !STRCMPI( pszName, "DEMPt" ) ||
		 !STRCMPI( pszName, "AdmPt" ) ||
		 !STRCMPI( pszName, "CommPt" ) ||
		 !STRCMPI( pszName, "SBAPt" ) ||
		 !STRCMPI( pszName, "Cntr" ) ||
		 !STRCMPI( pszName, "AdmBdry" ) ||
		 !STRCMPI( pszName, "CommBdry" ) ||
		 !STRCMPI( pszName, "SBBdry" ) ||
		 !STRCMPI( pszName, "RdASL" ) ||
		 !STRCMPI( pszName, "Cstline" ) ||
		 !STRCMPI( pszName, "WL" ) ||
		 !STRCMPI( pszName, "WStrL" ) ||
		 !STRCMPI( pszName, "WStrA" ) ||
		 !STRCMPI( pszName, "BldL" ) ||
		 !STRCMPI( pszName, "RdEdg" ) ||
		 !STRCMPI( pszName, "RdCompt" ) ||
		 !STRCMPI( pszName, "RailCL" ) ||
		 !STRCMPI( pszName, "RvrMgtBdry" ) ||
		 !STRCMPI( pszName, "LeveeEdge" ) ||
		 !STRCMPI( pszName, "RdMgtBdry" ) ||
		 !STRCMPI( pszName, "AdmArea" ) ||
		 !STRCMPI( pszName, "SBArea" ) ||
		 !STRCMPI( pszName, "WA" ) ||
		 !STRCMPI( pszName, "BldA" ) ||
		 !STRCMPI( pszName, "RdArea" ) ||
		 !STRCMPI( pszName, "DEM" ) )
	{
		m_nFlags |= TAG_ELEMENT;
		if ( m_validity == FGD_VALIDITY_UNKNOWN )
		{
			if ( !(m_nFlags & TAG_GI) || !(m_nFlags & TAG_DATASET) )
			{
				m_validity = FGD_VALIDITY_INVALID;
				return;
			}
			if ( STRCMPI( pszName, "DEM" ) )
			{
				m_validity = FGD_VALIDITY_INVALID;
				return;
			}
		}
	}
	else if ( !STRCMPI( pszName, "jps:EX_GeographicBoundingBox" ) )
	{
		m_nFlags |= TAG_BBOX;
	}
	else if ( !STRCMPI( pszName, "jps:westBoundLongitude" ) )
	{
		m_nFlags |= TAG_WEST;
	}
	else if ( !STRCMPI( pszName, "jps:eastBoundLongitude" ) )
	{
		m_nFlags |= TAG_EAST;
	}
	else if ( !STRCMPI( pszName, "jps:southBoundLatitude" ) )
	{
		m_nFlags |= TAG_SOUTH;
	}
	else if ( !STRCMPI( pszName, "jps:northBoundLatitude" ) )
	{
		m_nFlags |= TAG_NORTH;
	}
	else if ( !STRCMPI( pszName, "jps:low" ) )
	{
		m_nFlags |= TAG_LOW;
	}
	else if ( !STRCMPI( pszName, "jps:high" ) )
	{
		m_nFlags |= TAG_HIGH;
	}
	else if ( !STRCMPI( pszName, "jps:startSequence" ) )
	{
		m_nFlags |= TAG_STARTSEQ;
	}
	else if ( !STRCMPI( pszName, "jps:coordValues" ) )
	{
		m_nFlags |= TAG_COORDINATE;
	}
	else if ( !STRCMPI( pszName, "alti" ) )
	{
		m_bAltiReached = true;
	}

	m_szTempChar[0] = '\0';
	m_nTempChar = 0;
}


void GDALFGDReader::endElementValidateCbk( const char *pszName )
{
	char *pszStop;

	if ( !STRCMPI( pszName, "GI" ) )
	{
		m_nFlags &= ~TAG_GI;
	}
	else if ( !STRCMPI( pszName, "dataset" ) )
	{
		m_nFlags &= ~TAG_DATASET;
	}
	else if ( !STRCMPI( pszName, "GCP" ) ||
			  !STRCMPI( pszName, "ElevPt" ) ||
			  !STRCMPI( pszName, "DEMPt" ) ||
			  !STRCMPI( pszName, "AdmPt" ) ||
			  !STRCMPI( pszName, "CommPt" ) ||
			  !STRCMPI( pszName, "SBAPt" ) ||
			  !STRCMPI( pszName, "Cntr" ) ||
			  !STRCMPI( pszName, "AdmBdry" ) ||
			  !STRCMPI( pszName, "CommBdry" ) ||
			  !STRCMPI( pszName, "SBBdry" ) ||
			  !STRCMPI( pszName, "RdASL" ) ||
			  !STRCMPI( pszName, "Cstline" ) ||
			  !STRCMPI( pszName, "WL" ) ||
			  !STRCMPI( pszName, "WStrL" ) ||
			  !STRCMPI( pszName, "WStrA" ) ||
			  !STRCMPI( pszName, "BldL" ) ||
			  !STRCMPI( pszName, "RdEdg" ) ||
			  !STRCMPI( pszName, "RdCompt" ) ||
			  !STRCMPI( pszName, "RailCL" ) ||
			  !STRCMPI( pszName, "RvrMgtBdry" ) ||
			  !STRCMPI( pszName, "LeveeEdge" ) ||
			  !STRCMPI( pszName, "RdMgtBdry" ) ||
			  !STRCMPI( pszName, "AdmArea" ) ||
			  !STRCMPI( pszName, "SBArea" ) ||
			  !STRCMPI( pszName, "WA" ) ||
			  !STRCMPI( pszName, "BldA" ) ||
			  !STRCMPI( pszName, "RdArea" ) ||
			  !STRCMPI( pszName, "DEM" ) )
	{
		m_nFlags &= ~TAG_ELEMENT;
	}
	else if ( !STRCMPI( pszName, "jps:EX_GeographicBoundingBox" ) )
	{
		m_nFlags &= ~TAG_BBOX;
	}
	else if ( !STRCMPI( pszName, "jps:westBoundLongitude" ) )
	{
		m_dWest = CPLStrtod( m_szTempChar, &pszStop );
		m_nFlags &= ~TAG_WEST;
	}
	else if ( !STRCMPI( pszName, "jps:eastBoundLongitude" ) )
	{
		m_dEast = CPLStrtod( m_szTempChar, &pszStop );
		m_nFlags &= ~TAG_EAST;
	}
	else if ( !STRCMPI( pszName, "jps:southBoundLatitude" ) )
	{
		m_dSouth = CPLStrtod( m_szTempChar, &pszStop );
		m_nFlags &= ~TAG_SOUTH;
	}
	else if ( !STRCMPI( pszName, "jps:northBoundLatitude" ) )
	{
		m_dNorth = CPLStrtod( m_szTempChar, &pszStop );
		m_nFlags &= ~TAG_NORTH;
	}
	else if ( !STRCMPI( pszName, "jps:low" ) )
	{
		m_nFlags &= ~TAG_LOW;
	}
	else if ( !STRCMPI( pszName, "jps:high" ) )
	{
		m_nFlags &= ~TAG_HIGH;
	}
	else if ( !STRCMPI( pszName, "jps:startSequence" ) )
	{			
		m_nFlags &= ~TAG_STARTSEQ;
	}
	else if ( !STRCMPI( pszName, "jps:coordValues" ) )
	{
		if ( m_nFlags & TAG_HIGH )
			sscanf( m_szTempChar, "%d %d", &m_nHighX, &m_nHighY );
		else if ( m_nFlags & TAG_LOW )
			sscanf( m_szTempChar, "%d %d", &m_nLowX, &m_nLowY );
		else if ( m_nFlags & TAG_STARTSEQ )
			sscanf( m_szTempChar, "%d %d", &m_nStartX, &m_nStartY );

		m_nFlags &= ~TAG_COORDINATE;
	}
}


void GDALFGDReader::dataHandlerValidateCbk( const char *data, int nLen )
{
	if ( m_validity == FGD_VALIDITY_UNKNOWN &&
		( (m_nFlags & TAG_BBOX) && 
		((m_nFlags & TAG_WEST) || (m_nFlags & TAG_EAST) || (m_nFlags & TAG_SOUTH) || (m_nFlags & TAG_NORTH)) ) ||
		( ((m_nFlags & TAG_LOW) || (m_nFlags & TAG_HIGH) || (m_nFlags & TAG_STARTSEQ)) && (m_nFlags & TAG_COORDINATE) )
        )
	{
		char szTemp[MAX_LINES];
		CPLStrlcpy( szTemp, data, MAX_LINES );
		szTemp[nLen] = '\0';
		CPLStrlcat( m_szTempChar, szTemp, MAX_LINES );
		m_nTempChar += nLen;
	}
}

void GDALFGDReader::startElement(const char *pszName, const char **ppszAttr)
{
	m_szTempChar[0] = '\0';
	m_nTempChar = 0;

	if ( !STRCMPI( pszName, "alti" ) )
	{
		m_bAltiReached = true;
		if ( m_nPixelCount >= m_nStartPix )
		{
			m_nFlags |= TAG_ALTI;
		}
	}
	else if ( !STRCMPI( pszName, "GI" ) )
	{
		m_nFlags |= TAG_GI;
	}
	else if ( !STRCMPI( pszName, "dataset" ) )
	{
		m_nFlags |= TAG_DATASET;
	}
	else if ( !STRCMPI( pszName, "FGDFeature" ) ||
		 !STRCMPI( pszName, "GCP" ) ||
		 !STRCMPI( pszName, "ElevPt" ) ||
		 !STRCMPI( pszName, "DEMPt" ) ||
		 !STRCMPI( pszName, "AdmPt" ) ||
		 !STRCMPI( pszName, "CommPt" ) ||
		 !STRCMPI( pszName, "SBAPt" ) ||
		 !STRCMPI( pszName, "Cntr" ) ||
		 !STRCMPI( pszName, "AdmBdry" ) ||
		 !STRCMPI( pszName, "CommBdry" ) ||
		 !STRCMPI( pszName, "SBBdry" ) ||
		 !STRCMPI( pszName, "RdASL" ) ||
		 !STRCMPI( pszName, "Cstline" ) ||
		 !STRCMPI( pszName, "WL" ) ||
		 !STRCMPI( pszName, "WStrL" ) ||
		 !STRCMPI( pszName, "WStrA" ) ||
		 !STRCMPI( pszName, "BldL" ) ||
		 !STRCMPI( pszName, "RdEdg" ) ||
		 !STRCMPI( pszName, "RdCompt" ) ||
		 !STRCMPI( pszName, "RailCL" ) ||
		 !STRCMPI( pszName, "RvrMgtBdry" ) ||
		 !STRCMPI( pszName, "LeveeEdge" ) ||
		 !STRCMPI( pszName, "RdMgtBdry" ) ||
		 !STRCMPI( pszName, "AdmArea" ) ||
		 !STRCMPI( pszName, "SBArea" ) ||
		 !STRCMPI( pszName, "WA" ) ||
		 !STRCMPI( pszName, "BldA" ) ||
		 !STRCMPI( pszName, "RdArea" ) ||
		 !STRCMPI( pszName, "DEM" ) )
	{
		m_nFlags |= TAG_ELEMENT;
	}
}


void GDALFGDReader::endElement(const char *pszName)
{
	if ( !STRCMPI( pszName, "GI" ) )
	{
		m_nFlags &= ~TAG_GI;
	}
	else if ( !STRCMPI( pszName, "dataset" ) )
	{
		m_nFlags &= ~TAG_DATASET;
	}
	else if ( !STRCMPI( pszName, "GCP" ) ||
			  !STRCMPI( pszName, "ElevPt" ) ||
			  !STRCMPI( pszName, "DEMPt" ) ||
			  !STRCMPI( pszName, "AdmPt" ) ||
			  !STRCMPI( pszName, "CommPt" ) ||
			  !STRCMPI( pszName, "SBAPt" ) ||
			  !STRCMPI( pszName, "Cntr" ) ||
			  !STRCMPI( pszName, "AdmBdry" ) ||
			  !STRCMPI( pszName, "CommBdry" ) ||
			  !STRCMPI( pszName, "SBBdry" ) ||
			  !STRCMPI( pszName, "RdASL" ) ||
			  !STRCMPI( pszName, "Cstline" ) ||
			  !STRCMPI( pszName, "WL" ) ||
			  !STRCMPI( pszName, "WStrL" ) ||
			  !STRCMPI( pszName, "WStrA" ) ||
			  !STRCMPI( pszName, "BldL" ) ||
			  !STRCMPI( pszName, "RdEdg" ) ||
			  !STRCMPI( pszName, "RdCompt" ) ||
			  !STRCMPI( pszName, "RailCL" ) ||
			  !STRCMPI( pszName, "RvrMgtBdry" ) ||
			  !STRCMPI( pszName, "LeveeEdge" ) ||
			  !STRCMPI( pszName, "RdMgtBdry" ) ||
			  !STRCMPI( pszName, "AdmArea" ) ||
			  !STRCMPI( pszName, "SBArea" ) ||
			  !STRCMPI( pszName, "WA" ) ||
			  !STRCMPI( pszName, "BldA" ) ||
			  !STRCMPI( pszName, "RdArea" ) ||
			  !STRCMPI( pszName, "DEM" ) )
	{
		m_nFlags &= ~TAG_ELEMENT;
	}
	else
	{
		if ( !STRCMPI( pszName, "alti" ) )
		{
			if ( m_validity == FGD_VALIDITY_VALID && m_nPixelCount >= m_nStartPix )
			{
				float fPix = (float)CPLAtof( m_szTempChar );
				m_dqfPix.push_back( fPix );
			}
			m_nFlags &= ~TAG_ALTI;
			m_nPixelCount++;
		}

		m_szTempChar[0] = '\0';
		m_nTempChar = 0;
	}
}


void GDALFGDReader::dataHandler( const char *data, int nLen )
{
	if ( m_validity == FGD_VALIDITY_VALID && m_nFlags & TAG_ALTI )
	{
		char szTemp[MAX_LINES];
		CPLStrlcpy( szTemp, data, MAX_LINES );
		szTemp[nLen] = '\0';
		CPLStrlcat( m_szTempChar, szTemp, MAX_LINES );
		m_nTempChar += nLen;
	}
}


bool GDALFGDReader::parseNext()
{
	int nDone, nLen;
	char szBuf[BUFSIZ];
	XML_ParsingStatus status;

	if ( !m_Parser ) return false;

	XML_GetParsingStatus( m_Parser, &status );
	if ( status.parsing == XML_FINISHED )
		return false;

	{
		nLen = VSIFReadL( szBuf, 1, BUFSIZ, fp );
		nDone = VSIFEofL( fp );
		if ( XML_Parse( m_Parser, szBuf, nLen, nDone ) == XML_STATUS_ERROR )
		{
			fprintf(stderr, "Parse error at line %u:\n%s\n",
				XML_GetCurrentLineNumber( m_Parser ),
				XML_ErrorString(XML_GetErrorCode( m_Parser ) ) );
			return false;
		}
		if ( nDone )
			return false;
	}

	return true;

}

void GDALFGDReader::ResetReading()
{
	VSIFSeekL( fp, 0, SEEK_SET );
	m_nPixelsToPass = 0;
	m_nPixelCount = (m_nStartY * m_nCol) + m_nStartX;
	m_nPixelReturn = 0;
	m_nFlags = 0x0000;
	m_bAltiReached = false;

	if ( m_Parser )
		XML_ParserFree( m_Parser );

	m_Parser = XML_ParserCreate( NULL );

	XML_SetUserData( m_Parser, this );
	XML_SetElementHandler( m_Parser, ::startElement, ::endElement );
	XML_SetCharacterDataHandler( m_Parser, ::dataHandler );
	XML_SetUnknownEncodingHandler( m_Parser, XML_ShiftJISEncodingHandler, this );

	m_dqfPix.clear();
}


void GDALFGDReader::setLoi( int nLine )
{
	int nStartPix = (nLine * m_nCol);
	if ( nStartPix != m_nPixelReturn )
		m_dqfPix.clear();

	if ( nStartPix < m_nPixelReturn )
		ResetReading();

	m_nStartPix = nStartPix;
	m_nPixelReturn = nStartPix;
	m_nPixelsToPass = ( nLine == m_nStartY ) ? m_nStartX : ( nLine < m_nStartY ? m_nCol : 0 );

#if 1 
	if ( nLine == m_nStartY )
		int n = 0;
#endif
}


void GDALFGDReader::scanAll()
{
	// read to first ALTI tag
	while ( !m_bAltiReached )
	{
		if  ( !parseNext() )
		{
			m_validity = FGD_VALIDITY_INVALID;
			return;
		}
	}

	// check DEM information
	if ( m_nLowX == -1 || m_nLowY == -1 || m_nHighX == -1 || m_nHighY == -1 ||
		 m_dEast == FLT_MAX || m_dWest == FLT_MAX || m_dSouth == FLT_MAX || m_dNorth == FLT_MAX )
	{
		m_validity = FGD_VALIDITY_INVALID;
		return;
	}

	m_nRow = m_nHighY - m_nLowY + 1;
	m_nCol = m_nHighX - m_nLowX + 1;

	if ( m_nRow <= 0 || m_nCol <= 0 )
	{
		m_validity = FGD_VALIDITY_INVALID;
		return;
	}

	m_validity = FGD_VALIDITY_VALID;

	// read START SEQUENCE tag located near the end of file.
	if ( !readStartSequence() )
		m_nStartX = m_nStartY = 0;

	ResetReading();
}

bool GDALFGDReader::GetNextPix( float *pfPix )
{
	if ( m_nPixelsToPass )
	{
		*pfPix = FGD_NODATA;

		m_nPixelsToPass--;
		m_nPixelReturn++;
		return true;
	}

	while ( !m_dqfPix.size() )
	{
		if ( !parseNext() )
		{
			break;
		}
	}

	if ( !m_dqfPix.size() )
		return false;

	*pfPix = m_dqfPix.front();
	m_dqfPix.pop_front();
	m_nPixelReturn++;

	return true;
}


void GDALFGDReader::GetGeoTransform( double *padfTrans )
{
	padfTrans[0] = m_dWest;
	padfTrans[3] = m_dNorth;
	padfTrans[1] = (m_dEast - m_dWest) / m_nCol;
	padfTrans[2] = 0.0;
	padfTrans[4] = 0.0;
	padfTrans[5] = -1 * (m_dNorth - m_dSouth) / m_nRow;
}


bool GDALFGDReader::readStartSequence()
{
	char *pTagSeq, *pTagCoord;
	char szBuf[BUFSIZ+1], szData[20];

	VSIFSeekL( fp, -1*BUFSIZ, SEEK_END );
	VSIFReadL( szBuf, BUFSIZ, 1, fp );
	szBuf[BUFSIZ] = '\0';

	pTagSeq = strstr( szBuf, "<jps:startSequence>" );
	if ( !pTagSeq )
	{
		return false;
	}

	pTagSeq = strstr( pTagSeq, "<jps:coordValues>" );
	pTagSeq += 17;
	pTagCoord = strstr( pTagSeq, "</jps:coordValues>" );
	memcpy( szData, pTagSeq, pTagCoord - pTagSeq );
	szData[pTagCoord - pTagSeq] = '\0';

	sscanf( szData, "%d %d", &m_nStartX, &m_nStartY );

	return true;
}


#endif // #if defined(HAVE_EXPAT) && defined(HAVE_ICONV)
