
#include "cpl_string.h"
#ifdef HAVE_EXPAT
	#include <ogr_expat.h>
#endif
#include <ogr_core.h>
#include <ogr_feature.h>
#ifdef HAVE_ICONV
	#include <iconv.h>
#endif
#include <deque>

#define MAX_LINES 1000

#define FGD_NODATA -9999.0

#define	TAG_GI				0x0001	//	0000000000000001
#define	TAG_DATASET			0x0002	//	0000000000000010
#define	TAG_ELEMENT			0x0004	//	0000000000000100
#define	TAG_BBOX			0x0008	//	0000000000001000
#define	TAG_WEST			0x0010	//	0000000000010000
#define	TAG_EAST			0x0020	//	0000000000100000
#define	TAG_NORTH			0x0040	//	0000000001000000
#define	TAG_SOUTH			0x0080	//	0000000010000000
#define	TAG_LOW				0x0100	//	0000000100000000
#define	TAG_HIGH			0x0200	//	0000001000000000
#define	TAG_COORDINATE		0x0400	//	0000010000000000
#define	TAG_VALUES			0x0800	//	0000100000000000
#define	TAG_ALTI			0x1000	//	0001000000000000
#define	TAG_STARTSEQ		0x2000	//	0010000000000000

typedef enum
{
    FGD_VALIDITY_UNKNOWN,
    FGD_VALIDITY_INVALID,
    FGD_VALIDITY_VALID
} GDALFGDValidity;

class GDALFGDReader
{

public:
	GDALFGDReader( const char *pszFilename );
	~GDALFGDReader();

	bool isOpen() const { return m_bOpen; }

#if defined(HAVE_EXPAT) && defined(HAVE_ICONV)

	void startElement(const char *pszName, const char **ppszAttr);
	void startElementValidateCbk(const char *pszName, const char **ppszAttr);
	void endElement(const char *pszName);
	void endElementValidateCbk( const char *pszName );
	void dataHandler(const char *data, int nLen);
	void dataHandlerValidateCbk( const char *data, int nLen );
	int convertSJIStoUTF8( const char *s );

	inline GDALFGDValidity isValid() const { return m_validity; }
	void setLoi( int nLine );
	void ResetReading();
	bool GetNextPix( float *pfPix );
	void GetGeoTransform( double *padfTrans );
	inline int GetXSize() const { return m_nCol; }
	inline int GetYSize() const { return m_nRow; }

private:

	void scanAll();
	bool parseNext();
	bool readStartSequence();

	char				*m_pszVersion;
	GDALFGDValidity		m_validity;

	XML_Parser			m_Parser;
    int					m_nDataHandlerCounter;
	unsigned short		m_nFlags;

	VSILFILE *fp;
	iconv_t m_hIconv;

	char m_szTempChar[MAX_LINES];
	int m_nTempChar;

	int m_nLowX, m_nLowY;
	int m_nHighX, m_nHighY;
	int m_nRow, m_nCol;
	int m_nStartX, m_nStartY;
	bool m_bAltiReached;

	double m_dWest, m_dEast, m_dSouth, m_dNorth;

	int m_nStartPix;
	int m_nPixelsToPass;
	int m_nPixelCount;
	int m_nPixelReturn;

	std::deque<float> m_dqfPix;

#endif
	bool m_bOpen;
};
