
#include "cpl_string.h"
#ifdef HAVE_EXPAT
	#include <ogr_expat.h>
#endif
#include <ogr_core.h>
#include <ogr_feature.h>
#ifdef HAVE_ICONV
	#include <iconv.h>
#endif
#include <deque>
#include <vector>

#define MAX_LINES 1000

#define	TAG_GI				0x0001	//	0000000000000001
#define	TAG_DATASET			0x0002	//	0000000000000010
#define	TAG_ELEMENT			0x0004	//	0000000000000100
#define	TAG_ATTRIBUTE		0x0008	//	0000000000001000
#define	TAG_DATE_ATTRIBUTE	0x0010	//	0000000000010000
#define	TAG_DATE			0x0020	//	0000000000100000
#define	TAG_POS				0x0040	//	0000000001000000
#define	TAG_LOC				0x0080	//	0000000010000000
#define	TAG_AREA			0x0100	//	0000000100000000
#define	TAG_LINESTRING		0x0200	//	0000001000000000
#define	TAG_POLYGON			0x0400	//	0000010000000000
#define	TAG_EXTERIOR_RING	0x0800	//	0000100000000000
#define	TAG_INTERIOR_RING	0x1000	//	0001000000000000
#define	TAG_COORDINATE		0x2000	//	0010000000000000

typedef enum
{
    FGD_VALIDITY_UNKNOWN,
    FGD_VALIDITY_INVALID,
    FGD_VALIDITY_VALID
} OGRFGDValidity;

class FGDReader
{

public:
	FGDReader( const char *pszFilename );
	~FGDReader();

	bool isOpen() const { return m_bOpen; }
	inline OGRSpatialReference* GetSpatialRef(){ return m_poSRS; }

#if defined(HAVE_EXPAT) && defined(HAVE_ICONV)
	void startElement(const char *pszName, const char **ppszAttr);
	void endElement(const char *pszName);
	void dataHandler(const char *data, int nLen);
	int convertSJIStoUTF8( const char *s );

	OGRFeatureDefn *getLayerDefn() const { return m_poFeatureDefn; }
	OGRFGDValidity isValid() const { return m_validity; }

	void releaseLayerDefn() { m_poFeatureDefn->Release(); }

	OGRFeature* GetNextFeature();

	void ResetReading();

private:
	char				*m_pszVersion;
	OGRFGDValidity	m_validity;
	OGRFeatureDefn		*m_poFeatureDefn;

	XML_Parser			m_Parser;
    int					m_nDataHandlerCounter;
	unsigned short m_nFlags;

	VSILFILE *fp;
	iconv_t m_hIconv;

	bool parseNext();
	bool setLayerDefn( const char *pszName );

	char m_szTempChar[MAX_LINES];
	int m_nTempChar;

	std::deque<OGRFeature *> m_dqpoFeature;
	OGRFeature *m_poCurFeature;
	int m_nNextFID;

	OGRPoint *m_poPoint;
	OGRLineString *m_poLineString;
	OGRLinearRing *m_poRing;
	OGRPolygon *m_poPolygon;
	std::vector<OGRLinearRing *> m_vpoInteriorRings;
#endif

	bool m_bOpen;
	OGRSpatialReference *m_poSRS;
};
