
#include "ogr_fgd.h"

OGRFGDLayer::OGRFGDLayer( const char *pszFilename )
{
	nNextFID = 0;
	m_bValid = FALSE;

	poFeatureDefn = new OGRFeatureDefn( CPLGetBasename( pszFilename ) );
	poFeatureDefn->Reference();

#if defined(HAVE_EXPAT) && defined(HAVE_ICONV)
	m_pReader = new FGDReader( pszFilename );
	if ( m_pReader->isValid() != FGD_VALIDITY_VALID )
	{
		return;
	}

	m_bValid = TRUE;
	poFeatureDefn->SetGeomType( m_pReader->getLayerDefn()->GetGeomType() );
#endif
}


OGRFGDLayer::~OGRFGDLayer()
{
#if defined(HAVE_EXPAT) && defined(HAVE_ICONV)
	poFeatureDefn->Release();
	m_pReader->releaseLayerDefn();
	delete m_pReader;
#endif
}


OGRFeature* OGRFGDLayer::GetNextFeature()
{
#if defined(HAVE_EXPAT) && defined(HAVE_ICONV)
	while ( OGRFeature *poFeature = m_pReader->GetNextFeature() )
	{
		if ( ( m_poFilterGeom == NULL || FilterGeometry( poFeature->GetGeometryRef() ) ) &&
			 ( m_poAttrQuery == NULL || m_poAttrQuery->Evaluate( poFeature ) ) )
		{
			return poFeature;
		}

		delete poFeature;
	}
#endif

	return NULL;
}


OGRSpatialReference* OGRFGDLayer::GetSpatialRef()
{
	return m_pReader->GetSpatialRef();
}


OGRFeatureDefn* OGRFGDLayer::GetLayerDefn()
{
#if defined(HAVE_EXPAT) && defined(HAVE_ICONV)
	return m_pReader->getLayerDefn();
#else
	return NULL;
#endif
}


void OGRFGDLayer::ResetReading()
{
#if defined(HAVE_EXPAT) && defined(HAVE_ICONV)
	m_pReader->ResetReading();
#endif
}
